/*
  ==============================================================================

   This file is part of the JUCE examples.
   Copyright (c) 2020 - Raw Material Software Limited

   The code included in this file is provided under the terms of the ISC license
   http://www.isc.org/downloads/software-support-policy/isc-license. Permission
   To use, copy, modify, and/or distribute this software for any purpose with or
   without fee is hereby granted provided that the above copyright notice and
   this permission notice appear in all copies.

   THE SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY, AND ALL WARRANTIES,
   WHETHER EXPRESSED OR IMPLIED, INCLUDING MERCHANTABILITY AND FITNESS FOR
   PURPOSE, ARE DISCLAIMED.

  ==============================================================================
*/

#pragma once

using namespace dsp;

//==============================================================================
struct DSP_ParameterBase    : public ChangeBroadcaster
{
    DSP_ParameterBase (const String& labelName) : name (labelName) {}
    virtual ~DSP_ParameterBase() {}

    virtual Component* getComponent() = 0;

    virtual int getPreferredHeight()  = 0;
    virtual int getPreferredWidth()   = 0;

    String name;

    JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (DSP_ParameterBase)
};

//==============================================================================
struct SliderParameter   : public DSP_ParameterBase
{
    SliderParameter (Range<double> range, double skew, double initialValue,
                     const String& labelName, const String& suffix = {})
        : DSP_ParameterBase (labelName)
    {
        slider.setRange (range.getStart(), range.getEnd(), 0.01);
        slider.setSkewFactor (skew);
        slider.setValue (initialValue);
        slider.setColour (Slider::thumbColourId, Colour (0xffdbf1dc));
        //slider.setColour (Slider::thumbColourId, Colour (0xff797fed));
        //slider.setColour (Slider::textBoxTextColourId, Colour (0xff79ed7f));
        //slider.setColour (Slider::textBoxBackgroundColourId, Colour (0x004da7d3));
        //slider.setColour (Slider::textBoxOutlineColourId, Colour (0xff495358));
        //slider.setColour (Slider::trackColourId, Colour (0xff79ed7f));
        //slider.setColour (Slider::backgroundColourId, Colour (0xff28a2df));

        if (suffix.isNotEmpty())
            slider.setTextValueSuffix (suffix);

        slider.onValueChange = [this] { sendChangeMessage(); };
    }

    Component* getComponent() override    { return &slider; }

    int getPreferredHeight() override     { return 40; }
    int getPreferredWidth()  override     { return 500; }

    double getCurrentValue() const        { return slider.getValue(); }

private:
    Slider slider;
};

//==============================================================================
struct ChoiceParameter    : public DSP_ParameterBase
{
    ChoiceParameter (const StringArray& options, int initialId, const String& labelName)
        : DSP_ParameterBase (labelName)
    {
        parameterBox.addItemList (options, 1);
        parameterBox.onChange = [this] { sendChangeMessage(); };

        parameterBox.setSelectedId (initialId);
    }

    Component* getComponent() override    { return &parameterBox; }

    int getPreferredHeight() override     { return 25; }
    int getPreferredWidth()  override     { return 250; }

    int getCurrentSelectedID() const      { return parameterBox.getSelectedId(); }

private:
    ComboBox parameterBox;
};

//==============================================================================
class AudioThumbnailComponent    : public Component,
                                   public FileDragAndDropTarget,
                                   public ChangeBroadcaster,
                                   private ChangeListener,
                                   private Timer
{
public:
    AudioThumbnailComponent (AudioDeviceManager& adm, AudioFormatManager& afm)
        : audioDeviceManager (adm),
          thumbnailCache (5),
          thumbnail (128, afm, thumbnailCache)
    {
        thumbnail.addChangeListener (this);
    }

    ~AudioThumbnailComponent() override
    {
        thumbnail.removeChangeListener (this);
    }

    void paint (Graphics& g) override
    {
        g.fillAll (Colour (0xff495358));

        g.setColour (Colour (0xff91ec95));   //     Wave Colour

        if (thumbnail.getTotalLength() > 0.0)
        {
            thumbnail.drawChannels (g, getLocalBounds().reduced (2),
                                    0.0, thumbnail.getTotalLength(), 0.98f);

            g.setColour (Colours::black.withAlpha (0.10f));
            g.fillRect (static_cast<float> (currentPosition * getWidth()), 0.0f,
                        (getWidth()+1.0f), static_cast<float> (getHeight()));
        
        if (currentPosition == 0)
        {
            g.setColour (Colours::white.withAlpha (0.0f));
        }
        else
        {
            g.setColour (Colours::white.withAlpha (0.75f));
        }
            g.fillRect (static_cast<float> (currentPosition * getWidth()), 0.0f,
                        1.0f, static_cast<float> (getHeight()));
        }
        else
        {
            g.drawFittedText ("No audio file loaded.\nDrop a file here or click the \"Load File...\" button.", getLocalBounds(),
                              Justification::centred, 2);
            //g.drawFittedText ("No audio file loaded.\nDrop a file here or click the \"Load File...\" button.\nHot Keys: Stop - NumberPad0, Play - NumberPad1, Play-Pause - Space Key.", getLocalBounds(),
                              //Justification::centred, 2);
        }
    }

    bool isInterestedInFileDrag (const StringArray&) override          { return true; }
    void filesDropped (const StringArray& files, int, int) override    { loadURL (URL (File (files[0])), true); }

    void setCurrentURL (const URL& u)
    {
        if (currentURL == u)
            return;

        loadURL (u);
        //transportSource->start();   /// PLAY WHEN LOADING URL FILE /// MY CHANGE
    }

    URL getCurrentURL()    { return currentURL; }

    void setTransportSource (AudioTransportSource* newSource)
    {
        transportSource = newSource;

        struct ResetCallback  : public CallbackMessage
        {
            ResetCallback (AudioThumbnailComponent& o) : owner (o) {}
            void messageCallback() override    { owner.reset(); }

            AudioThumbnailComponent& owner;
        };

        (new ResetCallback (*this))->post();
    }

private:
    AudioDeviceManager& audioDeviceManager;
    AudioThumbnailCache thumbnailCache;
    AudioThumbnail thumbnail;
    AudioTransportSource* transportSource = nullptr;

    URL currentURL;
    double currentPosition = 0.0;

    //==============================================================================
    void changeListenerCallback (ChangeBroadcaster*) override    { repaint(); }

    void reset()
    {
        currentPosition = 0.0;
        repaint();

        if (transportSource == nullptr)
            stopTimer();
        else
            startTimerHz (25);
    }

    void loadURL (const URL& u, bool notify = false)
    {
        if (currentURL == u)
            return;

        currentURL = u;

        InputSource* inputSource = nullptr;

       #if ! JUCE_IOS
        if (u.isLocalFile())
        {
            inputSource = new FileInputSource (u.getLocalFile());
        }
        else
       #endif
        {
            if (inputSource == nullptr)
                inputSource = new URLInputSource (u);
        }

        thumbnail.setSource (inputSource);

        if (notify)
            sendChangeMessage();
    }

    void timerCallback() override
    {
        if (transportSource != nullptr)
        {
            currentPosition = transportSource->getCurrentPosition() / thumbnail.getTotalLength();
            repaint();
        }
    }
    
    //void mouseDrag or mouseDoubleClick

    void mouseDown (const MouseEvent& e) override
    {
        if (transportSource != nullptr)
        {
            const ScopedLock sl (audioDeviceManager.getAudioCallbackLock());

            transportSource->setPosition ((jmax (static_cast<double> (e.x), 0.0) / getWidth())
                                            * thumbnail.getTotalLength());
            transportSource->start();   // PAY WHEN MOUSE CLICK
        }
    }
    
    void mouseDrag (const MouseEvent& e) override
    {
        if (transportSource != nullptr)
        {
            const ScopedLock sl (audioDeviceManager.getAudioCallbackLock());

            transportSource->setPosition ((jmax (static_cast<double> (e.x), 0.0) / getWidth())
                                            * thumbnail.getTotalLength());
            transportSource->start();   // PAY WHEN MOUSE CLICK
        }
    }
};

//==============================================================================
class ProjectParametersComponent    : public Component
{
public:
    ProjectParametersComponent (const std::vector<DSP_ParameterBase*>& demoParams)
    {
        parameters = demoParams;

        for (auto demoParameter : parameters)
        {
            addAndMakeVisible (demoParameter->getComponent());

            auto* paramLabel = new Label ({}, demoParameter->name);

            paramLabel->attachToComponent (demoParameter->getComponent(), true);
            paramLabel->setJustificationType (Justification::centredLeft);
            addAndMakeVisible (paramLabel);
            labels.add (paramLabel);
        }
    }
    
    void paint (Graphics& g) override
    {
        //g.fillAll (Colour (0xff0c9de6));    //COLOUR OF CENTRE BODY
        //g.drawRoundedRectangle(0, 0, 400, 45, 4, 2);
    }

    void resized() override
    {
        auto bounds = getLocalBounds();
        bounds.removeFromLeft (100);

        for (auto* p : parameters)
        {
            auto* comp = p->getComponent();

            comp->setSize (jmin (bounds.getWidth(), p->getPreferredWidth()), p->getPreferredHeight());

            auto compBounds = bounds.removeFromTop (p->getPreferredHeight());
            comp->setCentrePosition (compBounds.getCentre());
        }
    }

    int getHeightNeeded()
    {
        auto height = 0;

        for (auto* p : parameters)
            height += p->getPreferredHeight();

        return height + 10;
    }

private:
    std::vector<DSP_ParameterBase*> parameters;
    OwnedArray<Label> labels;
};

//==============================================================================
template <class ProjectType>
struct DSPProject  : public AudioSource,
                  public ProcessorWrapper<ProjectType>,
                  private ChangeListener
{
    DSPProject (AudioSource& input)
        : inputSource (&input)
    {
        for (auto* p : getParameters())
            p->addChangeListener (this);
    }

    void prepareToPlay (int blockSize, double sampleRate) override
    {
        inputSource->prepareToPlay (blockSize, sampleRate);
        this->prepare ({ sampleRate, (uint32) blockSize, 2 });
    }

    void releaseResources() override
    {
        inputSource->releaseResources();
    }

    void getNextAudioBlock (const AudioSourceChannelInfo& bufferToFill) override
    {
        jassert (bufferToFill.buffer != nullptr);

        inputSource->getNextAudioBlock (bufferToFill);

        AudioBlock<float> block (*bufferToFill.buffer,
                                 (size_t) bufferToFill.startSample);

        ScopedLock audioLock (audioCallbackLock);
        this->process (ProcessContextReplacing<float> (block));
    }

    const std::vector<DSP_ParameterBase*>& getParameters()
    {
        return this->processor.parameters;
    }

    void changeListenerCallback (ChangeBroadcaster*) override
    {
        ScopedLock audioLock (audioCallbackLock);
        static_cast<ProjectType&> (this->processor).updateParameters();
    }

    CriticalSection audioCallbackLock;

    AudioSource* inputSource;
};

//==============================================================================
template <class ProjectType>
class AudioFileReaderComponent  : public Component,
                                  private TimeSliceThread,
                                  private Value::Listener,
                                  private ChangeListener
{
public:
    //==============================================================================
    AudioFileReaderComponent()
        : TimeSliceThread ("Audio File Reader Thread"),
          header (audioDeviceManager, formatManager, *this)
    {
        loopState.addListener (this);

        formatManager.registerBasicFormats();
        audioDeviceManager.addAudioCallback (&audioSourcePlayer);

       #ifndef JUCE_RUNNER
        audioDeviceManager.initialiseWithDefaultDevices (0, 2);
       #endif

        init();
        startThread();

        setOpaque (true);

        addAndMakeVisible (header);

        setSize (800, 250);
    }

    ~AudioFileReaderComponent() override
    {
        signalThreadShouldExit();
        stop();
        audioDeviceManager.removeAudioCallback (&audioSourcePlayer);
        waitForThreadToExit (10000);
    }

    void paint (Graphics& g) override       // BOTTOM BODY
    {
        g.setColour (getLookAndFeel().findColour (ResizableWindow::backgroundColourId));
        g.fillRect (getLocalBounds());
        //g.fillAll (Colour (0xff442222));
        //g.setColour (Colour (0xff442222));
    }

    void resized() override
    {
        auto r = getLocalBounds();

        header.setBounds (r.removeFromTop (120)); // WAVE BIG HEIGHT (220)

        r.removeFromTop (20);

        if (parametersComponent.get() != nullptr)
            parametersComponent->setBounds (r.removeFromTop (parametersComponent->getHeightNeeded()).reduced (20, 0));
    }

    //==============================================================================
    bool loadURL (const URL& fileToPlay)
    {
        stop();

        audioSourcePlayer.setSource (nullptr);
        getThumbnailComponent().setTransportSource (nullptr);
        transportSource.reset();
        readerSource.reset();

        AudioFormatReader* newReader = nullptr;

       #if ! JUCE_IOS
        if (fileToPlay.isLocalFile())
        {
            newReader = formatManager.createReaderFor (fileToPlay.getLocalFile());
        }
        else
       #endif
        {
            if (newReader == nullptr)
                newReader = formatManager.createReaderFor (fileToPlay.createInputStream (false));
        }

        reader.reset (newReader);

        if (reader.get() != nullptr)
        {
            readerSource.reset (new AudioFormatReaderSource (reader.get(), false));
            readerSource->setLooping (loopState.getValue());

            init();

            return true;
        }

        return false;
    }

     void togglePlay()
    {
        if (playState.getValue())
           { stop();
            playState = true;}
        else
            play();
    }
    
    
    void stop()
    {
        playState = false;

        if (transportSource.get() != nullptr)
        {
            transportSource->stop();
           // transportSource->setPosition (0);
        }
    }
    
    void toggleStop()
    {
        //playState = false;  MY CHANGE
        //playState = true;

        if (transportSource.get() != nullptr)
        {
            transportSource->stop();
            transportSource->setPosition (0);
        }
        if (playState.getValue() && ! transportSource->isPlaying())
                              //MY CHANGE
            playState = true;
    }
    
    bool keyPressed (const KeyPress& key) override
    {
                   
            if (key == KeyPress::spaceKey)
            {
            if (playState.getValue())
                { stop();
                playState = true;}
            else
                play();
            }
            
            if (key == KeyPress::numberPad0)
            {
                if (transportSource.get() != nullptr)
                {
                    transportSource->stop();
                    transportSource->setPosition (0);
                }
            }
            //if (key == KeyPress ('v', ModifierKeys::ctrlModifier, 0))            
            //if (key == KeyPress (KeyPress::spaceKey, ModifierKeys::shiftModifier, 0))
            //if (key == KeyPress ('v'))
            if (key == KeyPress::numberPad1)
            {
            
                if (readerSource.get() == nullptr)
                {
                    play();
                    transportSource->setPosition (0);
                }
                    else
                {
                    transportSource->setPosition (0);
                    play();
                }
                
            }    
            
            if (key == KeyPress::leftKey)
            {
                    transportSource->setPosition(transportSource->getCurrentPosition()-5);
                    
                if (transportSource->getCurrentPosition() <= 0)                
                {
                    transportSource->stop();
                    transportSource->setPosition (0);
                    //transportSource->stop();
                }             
            }
            
            if (key == KeyPress::rightKey)
            {
                                
                    transportSource->setPosition(transportSource->getCurrentPosition()+5);
                if (transportSource->getCurrentPosition() >= transportSource->getLengthInSeconds())
                {      
                    transportSource->setPosition (0);
                    play();
                    //transportSource->stop();
                }
            }
                   
            //if (key == KeyPress::leftKey)
            if (key == KeyPress (KeyPress::leftKey, ModifierKeys::commandModifier, 0))
            {
                    transportSource->setPosition(transportSource->getCurrentPosition()-15);
                    
                if (transportSource->getCurrentPosition() <= 0)                
                {
                    transportSource->stop();
                    transportSource->setPosition (0);
                    //transportSource->stop();
                }             
            }
            
            //if (key == KeyPress::rightKey)
            if (key == KeyPress (KeyPress::rightKey, ModifierKeys::commandModifier, 0))
            {
                                
                    transportSource->setPosition(transportSource->getCurrentPosition()+15);
                if (transportSource->getCurrentPosition() >= transportSource->getLengthInSeconds())
                {      
                    transportSource->setPosition (0);
                    play();
                    //transportSource->stop();
                }
            }
            
            if (key == KeyPress (KeyPress::leftKey, ModifierKeys::altModifier, 0))
            {
                    transportSource->setPosition(transportSource->getCurrentPosition()-25);
                    
                if (transportSource->getCurrentPosition() <= 0)                
                {
                    transportSource->stop();
                    transportSource->setPosition (0);
                    //transportSource->stop();
                }             
            }
            
            //if (key == KeyPress::rightKey)
            if (key == KeyPress (KeyPress::rightKey, ModifierKeys::altModifier, 0))
            {
                                
                    transportSource->setPosition(transportSource->getCurrentPosition()+25);
                if (transportSource->getCurrentPosition() >= transportSource->getLengthInSeconds())
                {      
                    transportSource->setPosition (0);
                    play();
                    //transportSource->stop();
                }
            }
                
            if (key == KeyPress::numberPad4)
            {
                transportSource->setPosition (60);
            }
            if (key == KeyPress::numberPad5)
            {
                transportSource->setPosition (80);         
            }  
            if (key == KeyPress::numberPad6)
            {
                transportSource->setPosition (100);         
            } 
            if (key == KeyPress::numberPad7)
            {
                transportSource->setPosition (120);         
            } 
            if (key == KeyPress::numberPad8)
            {
                transportSource->setPosition (140);         
            } 
            
            
    return true; 
    }

    void init()
    {
        if (transportSource.get() == nullptr)
        {
            transportSource.reset (new AudioTransportSource());
            transportSource->addChangeListener (this);

            if (readerSource.get() != nullptr)
            {
                if (auto* device = audioDeviceManager.getCurrentAudioDevice())
                {
                    transportSource->setSource (readerSource.get(), roundToInt (device->getCurrentSampleRate()), this, reader->sampleRate);

                    getThumbnailComponent().setTransportSource (transportSource.get());
                }
                transportSource->start(); // PLAY FILE WHEN DRAG IT TO WINDOW  // MY CHANGE
            }
        }

        audioSourcePlayer.setSource (nullptr);
        currentProject.reset();

        if (currentProject.get() == nullptr)
            currentProject.reset (new DSPProject<ProjectType> (*transportSource));

        audioSourcePlayer.setSource (currentProject.get());

        initParameters();
    }

    void play()
    {
        if (readerSource.get() == nullptr)
            return;

        if (transportSource->getCurrentPosition() >= transportSource->getLengthInSeconds()
             || transportSource->getCurrentPosition() < 0)
            transportSource->setPosition (0);

        transportSource->start();
        playState = true;
    }

    void setLooping (bool shouldLoop)
    {
        if (readerSource.get() != nullptr)
            readerSource->setLooping (shouldLoop);
    }

    AudioThumbnailComponent& getThumbnailComponent()    { return header.thumbnailComp; }

    void initParameters()
    {
        auto& parameters = currentProject->getParameters();

        parametersComponent.reset();

        if (parameters.size() > 0)
        {
            parametersComponent.reset (new ProjectParametersComponent (parameters));
            addAndMakeVisible (parametersComponent.get());
        }

        resized();
    }
    void openButtonClicked()    /////////////////////////////////////////////
    {
    }

private:
    //==============================================================================
    class AudioPlayerHeader     : public Component,
                                  private ChangeListener,
                                  private Value::Listener
    {
    public:
        AudioPlayerHeader (AudioDeviceManager& adm,
                           AudioFormatManager& afm,
                           AudioFileReaderComponent& afr)
            : thumbnailComp (adm, afm),
              audioFileReader (afr)
        {
            setOpaque (true);

            addAndMakeVisible (loadButton);
            addAndMakeVisible (playButton);
            addAndMakeVisible (stopButton);
            addAndMakeVisible (loopButton);

            playButton.setColour (TextButton::buttonColourId, Colour (0xff79ed7f));
            playButton.setColour (TextButton::textColourOffId, Colours::black);

            loadButton.setColour (TextButton::buttonColourId, Colour (0xff797fed));
            loadButton.setColour (TextButton::textColourOffId, Colours::black);

            loadButton.onClick = [this] { openFile(); };
            playButton.onClick = [this] { audioFileReader.togglePlay(); };
            stopButton.onClick = [this] { audioFileReader.toggleStop(); };
            stopButton.setColour (TextButton::textColourOffId, Colours::black);
            stopButton.setColour (TextButton::buttonColourId, Colour (0xffed797f));

            addAndMakeVisible (thumbnailComp);
            thumbnailComp.addChangeListener (this);

            audioFileReader.playState.addListener (this);
            loopButton.getToggleStateValue().referTo (audioFileReader.loopState);
        }

        ~AudioPlayerHeader() override
        {
            audioFileReader.playState.removeListener (this);
        }

        void paint (Graphics& g) override
        {
            g.setColour (getLookAndFeel().findColour (ResizableWindow::backgroundColourId).darker());
            g.fillRect (getLocalBounds());
        }

        void resized() override
        {
            auto bounds = getLocalBounds();

            auto buttonBounds = bounds.removeFromLeft (jmin (250, bounds.getWidth() / 4));
            auto loopBounds = buttonBounds.removeFromBottom (30);

            loadButton.setBounds (buttonBounds.removeFromTop (buttonBounds.getHeight() / 3));
            playButton.setBounds (buttonBounds.removeFromTop (buttonBounds.getHeight() / 1.9));
            stopButton.setBounds (buttonBounds.removeFromTop (buttonBounds.getHeight()));
            //loadButton.setBounds (0, 0, 90, 30);
            //playButton.setBounds (0, 30, 90, 30);
            //stopButton.setBounds (0, 60, 90, 30);

            loopButton.setSize (0, 25);
            loopButton.changeWidthToFitText();
            loopButton.setCentrePosition (loopBounds.getCentre());

            thumbnailComp.setBounds (bounds);
        }

        AudioThumbnailComponent thumbnailComp;

    private:
        //==============================================================================
        void openFile()
        {
            //audioFileReader.stop();

            if (fileChooser != nullptr)
                return;

            SafePointer<AudioPlayerHeader> safeThis (this);

            if (! RuntimePermissions::isGranted (RuntimePermissions::readExternalStorage))
            {
                RuntimePermissions::request (RuntimePermissions::readExternalStorage,
                                             [safeThis] (bool granted) mutable
                                             {
                                                 if (granted)
                                                     safeThis->openFile();
                                             });
                return;
            }

            fileChooser.reset (new FileChooser ("Select an audio file...", File(), "*.wav;*.mp3;*.aif"));

            fileChooser->launchAsync (FileBrowserComponent::openMode | FileBrowserComponent::canSelectFiles,
                                      [safeThis] (const FileChooser& fc) mutable
                                      {
                                          if (safeThis == nullptr)
                                              return;

                                          if (fc.getURLResults().size() > 0)
                                          {
                                              auto u = fc.getURLResult();

                                              if (! safeThis->audioFileReader.loadURL (u))
                                                  NativeMessageBox::showOkCancelBox (AlertWindow::WarningIcon, "Error loading file", "Unable to load audio file", nullptr, nullptr);
                                              else
                                                  safeThis->thumbnailComp.setCurrentURL (u);
                                          }

                                          safeThis->fileChooser = nullptr;
                                      }, nullptr);
        }

        void changeListenerCallback (ChangeBroadcaster*) override
        {
            if (audioFileReader.playState.getValue())
                audioFileReader.stop();

            audioFileReader.loadURL (thumbnailComp.getCurrentURL());
        }

        void valueChanged (Value& v) override
        {
            playButton.setButtonText (v.getValue() ? "Pause" : "Play");
            playButton.setColour (TextButton::buttonColourId, v.getValue() ? Colour (0xff79ed7f) : Colour (0xff79ed7f));
        }

        //==============================================================================
        TextButton loadButton { "Load File..." }, playButton { "Play" }, stopButton { "Stop" };
        ToggleButton loopButton { "Loop File" };

        AudioFileReaderComponent& audioFileReader;
        std::unique_ptr<FileChooser> fileChooser;
  
           
};

    //==============================================================================
    void valueChanged (Value& v) override
    {
        if (readerSource.get() != nullptr)
            readerSource->setLooping (v.getValue());
    }

    void changeListenerCallback (ChangeBroadcaster*) override
    {
        if (playState.getValue() && ! transportSource->isPlaying())
            stop();
            else                    //MY CHANGE
            playState = true;
    }

    //==============================================================================
    // if this PIP is running inside the demo runner, we'll use the shared device manager instead
   #ifndef JUCE_RUNNER
    AudioDeviceManager audioDeviceManager;
   #else
    AudioDeviceManager& audioDeviceManager { getSharedAudioDeviceManager (0, 2) };
   #endif

    AudioFormatManager formatManager;
    Value playState { var (false) };
    Value loopState { var (false) };

    double currentSampleRate = 44100.0;
    uint32 currentBlockSize = 512;
    uint32 currentNumChannels = 2;

    std::unique_ptr<AudioFormatReader> reader;
    std::unique_ptr<AudioFormatReaderSource> readerSource;
    std::unique_ptr<AudioTransportSource> transportSource;
    std::unique_ptr<DSPProject<ProjectType>> currentProject;

    AudioSourcePlayer audioSourcePlayer;

    AudioPlayerHeader header;

    AudioBuffer<float> fileReadBuffer;

    std::unique_ptr<ProjectParametersComponent> parametersComponent;
};
